(*
Author:  Christian Sternagel <c.sternagel@gmail.com> (2014, 2015)
Author:  René Thiemann <rene.thiemann@uibk.ac.at> (2015)
License: LGPL (see file COPYING.LESSER)
*)

section \<open>Peak Decreasingness\<close>

theory Peak_Decreasingness
imports
  "$AFP/Abstract-Rewriting/Abstract_Rewriting"
  "$AFP/Well_Quasi_Orders/Multiset_Extension"
begin

lemma SN_mulex: "SN {(x, y). P x y} \<Longrightarrow> SN {(x, y). mulex P\<inverse>\<inverse> y x}"
  using wfp_on_mulex_on_multisets [of "P\<inverse>\<inverse>" UNIV]
  by (auto simp: wfP_def SN_iff_wf converse_def)

text \<open>ARSs whose relation is the union of an indexed set of relations.\<close>
locale ars_labeled =
  fixes R :: "'b rel"
    and r :: "'a \<Rightarrow> 'b rel"
    and I :: "'a set"
  assumes R: "R = (\<Union>a\<in>I. r a)"
begin

inductive conv where
  empty [simp, intro!]: "conv {#} x x" |
  step [intro]: "a \<in> I \<Longrightarrow> (x, y) \<in> (r a)\<^sup>\<leftrightarrow> \<Longrightarrow> conv M y z \<Longrightarrow> conv (M + {#a#}) x z"

lemma conversion_conv_iff:
  "(x, y) \<in> R\<^sup>\<leftrightarrow>\<^sup>* \<longleftrightarrow> (\<exists>M. conv M x y)"
proof
  assume "(x, y) \<in> R\<^sup>\<leftrightarrow>\<^sup>*"
  then have "(x, y) \<in> (R \<union> R\<inverse>)\<^sup>*" by auto
  then show "\<exists>M. conv M x y"
    by (induct rule: converse_rtrancl_induct) (auto simp: R)
next
  assume "\<exists>M. conv M x y"
  then obtain M where "conv M x y" by auto
  then show "(x, y) \<in> R\<^sup>\<leftrightarrow>\<^sup>*"
    by (induct) (auto simp: conversion_def R intro: converse_rtrancl_into_rtrancl)
qed

lemma conv_singleton [intro]:
  "a \<in> I \<Longrightarrow> (x, y) \<in> (r a)\<^sup>\<leftrightarrow> \<Longrightarrow> conv {#a#} x y"
  by (metis conv.empty conv.step empty_neutral(1))

lemma conv_union_left [simp]:
  assumes "a \<in> I" and "(y, x) \<in> r a" and "conv (M + N) y z"
  shows "conv (M + ({#a#} + N)) x z"
proof -
  from assms have "conv ((M + N) + {#a#}) x z" by blast
  then show ?thesis by (simp add: ac_simps)
qed

lemma conv_union_right [simp]:
  assumes "a \<in> I" and "(x, y) \<in> (r a)" and "conv (M + N) y z"
  shows "conv (M + ({#a#} + N)) x z"
proof -
  from assms have "conv ((M + N) + {#a#}) x z" by blast
  then show ?thesis by (simp add: ac_simps)
qed

lemma conv_step' [intro]:
  assumes "a \<in> I" and "(y, z) \<in> (r a)\<^sup>\<leftrightarrow>" and "conv M x y"
  shows "conv (M + {#a#}) x z"
  using assms(3, 2, 1) by (induct) (auto simp: ac_simps)

lemma conv_trans:
  assumes "conv M x y" and "conv N y z"
  shows "conv (M + N) x z"
  using assms(2, 1) by (induct arbitrary: M) (auto simp: ac_simps)

lemma conv_commute:
  assumes "conv M x y" shows "conv M y x"
  using assms by (induct) auto

inductive seql where
  empty [simp, intro!]: "seql {#} x x" |
  step [intro]: "a \<in> I \<Longrightarrow> (z, y) \<in> r a \<Longrightarrow> seql M x y \<Longrightarrow> seql (M + {#a#}) x z"

lemma rtrancl_seql_iff:
  "(x, y) \<in> R\<^sup>* \<longleftrightarrow> (\<exists>M. seql M y x)"
proof
  assume "(x, y) \<in> R\<^sup>*"
  then show "\<exists>M. seql M y x"
    by (induct rule: converse_rtrancl_induct) (auto simp: R)
next
  assume "\<exists>M. seql M y x"
  then obtain M where "seql M y x" ..
  then show "(x, y) \<in> R\<^sup>*"
    by (induct) (auto simp: R rtrancl_converse intro: converse_rtrancl_into_rtrancl)
qed

lemma seql_singleton [intro]:
  "a \<in> I \<Longrightarrow> (x, y) \<in> r a \<Longrightarrow> seql {#a#} y x"
  by (metis empty_neutral(1) seql.empty seql.step)

lemma seql_imp_conv:
  assumes "seql M x y" shows "conv M x y"
  using assms by (induct) auto

lemma seql_union [simp]:
  "a \<in> I \<Longrightarrow> (z, y) \<in> r a \<Longrightarrow> seql (M + N) x y \<Longrightarrow> seql (M + ({#a#} + N)) x z"
  by (auto simp: ac_simps) (metis seql.step union_assoc)

lemma seql_left' [intro]:
  assumes "a \<in> I" and "(y, x) \<in> r a" and "seql M y z"
  shows "seql (M + {#a#}) x z"
  using assms(3, 2, 1) by (induct) (auto simp: ac_simps)

inductive valley where
  seql [simp, intro]: "seql M x y \<Longrightarrow> valley M x y" |
  left [intro]: "\<lbrakk>a \<in> I; (x, y) \<in> r a; valley M y z\<rbrakk> \<Longrightarrow> valley (M + {#a#}) x z"

lemma join_valley_iff:
  "(x, y) \<in> R\<^sup>\<down> \<longleftrightarrow> (\<exists>M. valley M x y)"
proof
  assume "(x, y) \<in> R\<^sup>\<down>"
  then obtain z where "(x, z) \<in> R\<^sup>*" and "(y, z) \<in> R\<^sup>*" by auto
  then obtain N where "(x, z) \<in> R\<^sup>*" and "seql N z y" by (auto simp: rtrancl_seql_iff)
  then show "\<exists>M. valley M x y"
    by (induct arbitrary: y rule: converse_rtrancl_induct)
       (auto simp: R dest: valley.left)
next
  assume "\<exists>M. valley M x y"
  then obtain M where "valley M x y" by auto
  then show "(x, y) \<in> R\<^sup>\<down>"
    by (induct) (auto iff: rtrancl_seql_iff intro: rtrancl_join_join)
qed

lemma valley_imp_conv:
  assumes "valley M x y" shows "conv M x y"
  using assms by (induct) (auto intro: seql_imp_conv)

text \<open>A conversion is either a valley or contains a peak.\<close>
lemma conv_valley_peak_cases:
  assumes "conv M x y"
  obtains
    (valley) "valley M x y" |
    (peak) K L a b t u v
    where "M = K + L + {#a, b#}"
      and "conv K x t"
      and "a \<in> I" and "(u, t) \<in> r a"
      and "b \<in> I" and "(u, v) \<in> r b"
      and "conv L v y"
using assms
proof (induct)
  case (empty x) then show ?case by auto
next
  case (step a x y M z)
  show ?case
  proof (rule step.hyps)
    assume *: "valley M y z"
    from \<open>(x, y) \<in> (r a)\<^sup>\<leftrightarrow>\<close> show ?thesis
    proof
      assume "(x, y) \<in> r a"
      with * and \<open>a \<in> I\<close> have "valley (M + {#a#}) x z" by auto
      then show ?thesis by fact
    next
      assume "(x, y) \<in> (r a)\<inverse>"
      then have "(y, x) \<in> r a" by simp
      from * show ?thesis
      proof (cases)
        assume "seql M y z"
        then have "valley (M + {#a#}) x z" using step by blast
        then show ?thesis by fact
      next
        fix b u N
        assume "b \<in> I" and "(y, u) \<in> r b"
          and [simp]: "M = N + {#b#}"
          and "valley N u z"
        then have *: "M - {#b#} = N" by simp
        have "M + {#a#} = {#} + (M - {#b#}) + {#a, b#}" by (simp add: ac_simps)
        moreover have "conv {#} x x" by blast
        moreover have "a \<in> I" and "(y, x) \<in> r a" by fact+
        moreover have "b \<in> I" and "(y, u) \<in> r b" by fact+
        moreover have "conv (M - {#b#}) u z"
          unfolding * by (rule valley_imp_conv) fact
        ultimately show ?thesis by fact
      qed
    qed
  next
    fix K L b c t u v
    assume [simp]: "M = K + L + {#b, c#}"
      and "conv K y t" and "b \<in> I" and "(u, t) \<in> r b"
      and "c \<in> I" and "(u, v) \<in> r c" and "conv L v z"
    moreover with step have "conv (K + {#a#}) x t" by blast
    moreover have "M + {#a#} = (K + {#a#}) + L + {#b, c#}" by (auto simp: ac_simps)
    show ?thesis by (rule step.prems) fact+
  qed
qed

end

text \<open>ARSs as unions of relations over an index set equipped with a well-founded order.\<close>
locale ars_labeled_sn = ars_labeled +
  fixes less :: "'b \<Rightarrow> 'b \<Rightarrow> bool" (infix "\<succ>" 50)
  assumes SN_less: "SN {(x, y). x \<succ> y}"
begin

definition mless :: "'b multiset \<Rightarrow> 'b multiset \<Rightarrow> bool" (infix "\<succ>\<^sub>M" 50) where
  "x \<succ>\<^sub>M y \<longleftrightarrow> mulex (op \<succ>)\<inverse>\<inverse> y x"

abbreviation mlesseq (infix "\<succeq>\<^sub>M" 50) where
  "x \<succeq>\<^sub>M y \<equiv> mless\<^sup>=\<^sup>= x y"

lemma SN_mless: "SN {(x, y). x \<succ>\<^sub>M y}"
  using SN_mulex [OF SN_less] by (auto simp: SN_defs mless_def)

lemma mless_mono:
  "M \<succ>\<^sub>M N \<Longrightarrow> K + M \<succ>\<^sub>M K + N"
  unfolding mless_def by (metis UNIV_I mulex_on_union multisets_UNIV)

lemma mless_trans:
  "as \<succ>\<^sub>M bs \<Longrightarrow> bs \<succ>\<^sub>M cs \<Longrightarrow> as \<succ>\<^sub>M cs"
  unfolding mless_def by (metis mulex_on_trans)

lemma mless_empty [simp]:
  "M \<noteq> {#} \<Longrightarrow>  M \<succ>\<^sub>M {#}"
  by (auto simp: mless_def)

lemma mless_all_strict:
  "M \<noteq> {#} \<Longrightarrow> \<forall>y. y \<in># N \<longrightarrow> (\<exists>x. x \<in># M \<and> x \<succ> y) \<Longrightarrow> M \<succ>\<^sub>M N"
  using mulex_on_all_strict [of M UNIV N "op \<succ>\<inverse>\<inverse>"] by (simp add: mless_def)

abbreviation downset2 :: "'b \<Rightarrow> 'b \<Rightarrow> 'b set" ("\<or>'(_,/ _')")
where
  "\<or>(a, b) \<equiv> {c \<in> I. a \<succ> c \<or> b \<succ> c}"

lemma downset2_multiset:
  assumes "\<forall>c. c \<in># M \<longrightarrow> a \<succ> c \<or> b \<succ> c"
  shows "{#a, b#} \<succ>\<^sub>M M"
  using assms
  by (metis mless_all_strict multi_member_last multi_member_skip multi_member_this single_not_empty union_eq_empty)

lemma Union_downset2_trans:
  "(s, t) \<in> (\<Union>c\<in>\<or>(a, b). (r c))\<^sup>\<leftrightarrow>\<^sup>* \<Longrightarrow>
   (t, u) \<in> (\<Union>c\<in>\<or>(a, b). (r c))\<^sup>\<leftrightarrow>\<^sup>* \<Longrightarrow>
   (s, u) \<in> (\<Union>c\<in>\<or>(a, b). (r c))\<^sup>\<leftrightarrow>\<^sup>*"
  by (auto simp: conversion_def)

lemma sym_step_in_downset2:
  assumes "(s, t) \<in> (r c)\<^sup>\<leftrightarrow>" and "c \<in> I" and "a \<succ> c \<or> b \<succ> c"
  shows "(s, t) \<in> (\<Union>c\<in>\<or>(a, b). r c)\<^sup>\<leftrightarrow>\<^sup>*"
  using assms by (auto)

lemma sym_steps_in_downset2:
  assumes "(s, t) \<in> ((r c)\<^sup>\<leftrightarrow>)\<^sup>*" and "c \<in> I" and "a \<succ> c \<or> b \<succ> c"
  shows "(s, t) \<in> (\<Union>c\<in>\<or>(a, b). r c)\<^sup>\<leftrightarrow>\<^sup>*"
  using assms
  by (induct s t, simp) (metis (mono_tags) Union_downset2_trans sym_step_in_downset2)

lemma conversion_in_downset2:
  assumes "(s, t) \<in> (r c)\<^sup>\<leftrightarrow>\<^sup>*" and "c \<in> I" and "a \<succ> c \<or> b \<succ> c"
  shows "(s, t) \<in> (\<Union>c\<in>\<or>(a, b). r c)\<^sup>\<leftrightarrow>\<^sup>*"
  using sym_steps_in_downset2 [OF assms [unfolded conversion_def]] .

lemma downset2_conversion_imp_conv:
  assumes "(t, u) \<in> (\<Union>c\<in>\<or>(a, b). r c)\<^sup>\<leftrightarrow>\<^sup>*"
  shows "\<exists>M. {#a, b#} \<succ>\<^sub>M M \<and> conv M t u"
proof -
  from assms [unfolded conversion_def]
    have "\<exists>M. (\<forall>c. c \<in># M \<longrightarrow> a \<succ> c \<or> b \<succ> c) \<and> conv M t u"
  proof (induct)
    case base
    show ?case
      by (metis conv.empty count_empty less_not_refl)
  next
    case (step u v)
    then obtain M where *: "\<forall>c. c \<in># M \<longrightarrow> a \<succ> c \<or> b \<succ> c" and "conv M t u" by blast
    from step(2) obtain c where "c \<in> I" and "a \<succ> c \<or> b \<succ> c"
      and "(u, v) \<in> (r c)\<^sup>\<leftrightarrow>" by auto
    with \<open>conv M t u\<close> have "conv (M + {#c#}) t v" by (metis conv_step')
    moreover have "\<forall>d. d \<in># M + {#c#} \<longrightarrow> a \<succ> d \<or> b \<succ> d"
      using * and \<open>a \<succ> c \<or> b \<succ> c\<close> by (auto)
    ultimately show ?case by blast
  qed
  with downset2_multiset [of _ a b]
    show ?thesis by blast
qed

end

locale ars_peak_decreasing = ars_labeled_sn +
  assumes peak_decreasing:
    "\<lbrakk>a \<in> I; b \<in> I; (x, y) \<in> r a; (x, z) \<in> r b\<rbrakk> \<Longrightarrow> (y, z) \<in> (\<Union>c\<in>\<or>(a, b). r c)\<^sup>\<leftrightarrow>\<^sup>*"
begin

lemma conv_join_mless_convD:
  assumes "conv M x y"
  shows "valley M x y \<or> (\<exists>N. M \<succ>\<^sub>M N \<and> conv N x y)"
using assms
proof (cases rule: conv_valley_peak_cases)
  case (peak K L a b t u v)
  with peak_decreasing
    have "(t, v) \<in> (\<Union>c\<in>\<or>(a, b). r c)\<^sup>\<leftrightarrow>\<^sup>*" by blast
  from downset2_conversion_imp_conv [OF this] obtain N
    where "{#a, b#} \<succ>\<^sub>M N" and "conv N t v" by blast
  then have "M \<succ>\<^sub>M K + N + L"
    unfolding peak by (auto simp: ac_simps intro: mless_mono)
  moreover have "conv (K + N + L) x y"
    using peak and \<open>conv N t v\<close> by (blast intro: conv_trans)
  ultimately show ?thesis by blast
qed simp

text \<open>A conversion is either a valley or can be made smaller w.r.t.\ @{const mless}.\<close>
lemma conv_join_mless_conv_cases:
  assumes "conv M x y"
  obtains
    (valley) "valley M x y" |
    (conv) N
    where "M \<succ>\<^sub>M N" and "conv N x y"
  using conv_join_mless_convD [OF assms] by blast

lemma conv_imp_valley:
  assumes "conv M x y"
  shows "\<exists>N. M \<succeq>\<^sub>M N \<and> valley N x y"
using SN_mless and assms
proof (induct)
  case (IH M)
  from \<open>conv M x y\<close> show ?case
  proof (cases rule: conv_join_mless_conv_cases)
    case valley then show ?thesis by auto
  next
    case (conv N)
    with IH(1) [of N] show ?thesis by (auto elim: mless_trans)
  qed
qed

text \<open>Every peak decreasing ARS is confluent.\<close>
lemma CR: "CR R"
proof
  fix x y z
  assume "x \<in> UNIV" and "(x, y) \<in> R\<^sup>*" and "(x, z) \<in> R\<^sup>*"
  then have "(y, z) \<in> R\<^sup>\<leftrightarrow>\<^sup>*" by (metis in_mono meetI meet_imp_conversion)
  then obtain M where "conv M y z" by (auto simp: conversion_conv_iff)
  from conv_imp_valley [OF this]
    show "(y, z) \<in> R\<^sup>\<down>" by (auto iff: join_valley_iff)
qed

end

context ars_labeled_sn
begin

lemma mset_replicate_Suc:
  "(\<forall>x. x \<in># mset (replicate (Suc n) y) \<longrightarrow> P x) \<longleftrightarrow> P y"
  by (induct n) auto

lemma same_label_conversion_imp_conv:
  assumes "(t, u) \<in> (r c)\<^sup>\<leftrightarrow>\<^sup>*" and "c \<in> I" and "a \<succ> c \<or> b \<succ> c"
  shows "\<exists>M. {#a, b#} \<succ>\<^sub>M M \<and> conv M t u"
proof -
  let ?M = "\<lambda>n. mset (replicate n c)"

  have "\<And>n. \<forall>y. y \<in># ?M (Suc n) \<longrightarrow> (\<exists>x. x \<in># {#a, b#} \<and> x \<succ> y)"
    unfolding mset_replicate_Suc using \<open>a \<succ> c \<or> b \<succ> c\<close> by auto
  from mulex_on_all_strict [OF _ _ _ this, of UNIV]
    have mless: "\<And>n. {#a, b#} \<succ>\<^sub>M ?M (Suc n)"
    by (auto simp: mless_def intro: mulex_on_mono)

  from assms(1) have "(t, u) \<in> (r c \<union> (r c)\<inverse>)\<^sup>*" by auto
  then have "\<exists>n. {#a, b#} \<succ>\<^sub>M ?M n \<and> conv (?M n) t u"
  proof (induct)
    case base
    have "{#a, b#} \<succ>\<^sub>M ?M 0" by (simp add: mless_def)
    moreover have "conv (?M 0) t t" by simp
    ultimately show ?case by blast
  next
    case (step y z)
    then obtain n where "conv (?M n) t y" by blast
    moreover from \<open>(y, z) \<in> r c \<union> (r c)\<inverse>\<close> and \<open>c \<in> I\<close>
      have "conv {#c#} y z" by (auto)
    ultimately have "conv (?M n + {#c#}) t z" by (rule conv_trans)
    then have "conv (?M (Suc n)) t z" by simp
    with mless show ?case by blast
  qed
  then show ?thesis by blast
qed

lemma ars_peak_decreasing_alt_intro:
  assumes *: "\<And>a b x y z.
    \<lbrakk>a \<in> I; b \<in> I; (x, y) \<in> r a; (x, z) \<in> r b\<rbrakk> \<Longrightarrow>
    \<exists>c. c \<in> I \<and> (a \<succ> c \<or> b \<succ> c) \<and> (y, z) \<in> (r c)\<^sup>\<leftrightarrow>\<^sup>*"
  shows "ars_peak_decreasing R r I less"
proof (unfold_locales)
  fix a b x y z
  assume "a \<in> I" and "b \<in> I" and "(x, y) \<in> r a" and "(x, z) \<in> r b"
  from * [OF this] obtain c
    where "(y, z) \<in> (r c)\<^sup>\<leftrightarrow>\<^sup>*" and "c \<in> I" and "a \<succ> c \<or> b \<succ> c" by blast
  from conversion_in_downset2 [OF this]
    show "(y, z) \<in> (\<Union>c\<in>\<or>(a, b). r c)\<^sup>\<leftrightarrow>\<^sup>*" .
qed

end

text \<open>Partitioning of a relation that labels steps by their source.\<close>
definition source_step :: "'a rel \<Rightarrow> 'a \<Rightarrow> 'a rel" where
  "source_step R x = {(x, y) | y. (x, y) \<in> R}"

lemma source_step_label [iff]:
  "(x, y) \<in> source_step R z \<longleftrightarrow> (x, y) \<in> R \<and> z = x"
  by (auto simp: source_step_def)

end

